import copy
import StringIO
import time
import zlib
from collections import OrderedDict
from datetime import datetime

from os_rotatefile import open_file

from default_schema import META_SCHEMA
from validator import TIME_FORMAT, create_validator


class RecordProcessor(object):
    def process(self, record, **kwargs):
        return record


class SpageRecordProcessor(RecordProcessor):
    def __init__(self, validator, compress=True):
        self._validator = validator
        self._compress = compress

    def _compress_data(self, data):
        return zlib.compress(data)

    def process(self, record, **kwargs):
        if not record['http_header']:
            record.pop('http_header')
        inner_header = record['inner_header']
        store_size = original_size = len(
            record['data']) if record['data'] is not None else -1
        inner_header['Original-Size'] = original_size

        if original_size >= 0:
            data = record['data']
            if inner_header.get('Type', None) is None:
                if self._compress:
                    inner_header['Type'] = 'compressed'
                    data = self._compress_data(data)
                    store_size = len(data)
                else:
                    inner_header['Type'] = 'flat'

            record['data'] = data
            inner_header['Store-Size'] = store_size
        else:
            for i in ['Original-Size', 'Store-Size']:
                if i in inner_header:
                    inner_header.pop(i)
            record.pop('data')

        self._validator.validate(record)
        return record


class RecordEncoder(object):
    def dumps(self, record, **kwargs):
        raise NotImplementedError


class SpageRecordEncoder(RecordEncoder):

    def __init__(self, allowed_inner_header_keys=None):
        self._inner_header_keys = allowed_inner_header_keys

    def _http_header_str(self, http_header):
        if not http_header:
            return None
        return '\r\n'.join([': '.join((k.strip(), v.strip())) for k, v in http_header.items()])

    def _inner_header_str(self, inner_header):
        o = StringIO.StringIO()
        keys = self._inner_header_keys if self._inner_header_keys \
            else inner_header.keys()
        for k in keys:
            if k in inner_header:
                v = inner_header[k]
                if v is None:
                    continue
                if isinstance(v, datetime):
                    v = v.strftime(TIME_FORMAT)
                o.write(': '.join((str(k).strip(), str(v).strip())))
                o.write('\n')

        data_length = o.tell() - 1
        o.seek(0)
        return o.read(data_length)

    def dumps(self, record, **kwargs):
        o = StringIO.StringIO()
        o.write(record['url'])
        o.write('\n')

        inner_header_str = self._inner_header_str(record['inner_header'])
        o.write(inner_header_str)
        o.write('\n\n')

        http_header_str = self._http_header_str(
            record.get('http_header', None))
        if not http_header_str:
            o.write('\r\n')
        else:
            o.write(http_header_str)
            o.write('\r\n\r\n')

        data = record.get("data", None)
        if data is not None:
            o.write(data)
            o.write('\r\n')

        o.seek(0)
        return o.read()


class RecordWriter(object):
    def write(self, f, url, inner_header=None, http_header=None, data=None):
        raise NotImplementedError


class SpageRecordWriter(RecordWriter):
    def __init__(self, processor, encoder):
        self._processor = processor
        self._encoder = encoder

    def write(self, f, url, inner_header=None, http_header=None, data=None):
        record = {}
        record["url"] = url
        record["inner_header"] = {} if inner_header is None else copy.deepcopy(
            inner_header)
        record["http_header"] = {} if http_header is None else copy.deepcopy(
            http_header)
        record["data"] = data

        record = self._processor.process(record)
        f.write(self._encoder.dumps(record))


def create_writer(**kwargs):
    validator = kwargs.get('validator', None)
    if validator is None:
        validator = create_validator(
            META_SCHEMA) if validator is None else validator
    processor = SpageRecordProcessor(validator, kwargs.get('compress', True))
    allowed_keys = validator.schema['properties']['inner_header']['properties'].keys(
    )
    encoder = SpageRecordEncoder(allowed_keys)
    return SpageRecordWriter(processor, encoder)


class SpageWriter(object):
    def __init__(self, base_filename, roll_size='1G', compress=True, validator=None):
        self._fp = open_file(base_filename, 'w', roll_size=roll_size)
        self._record_writer = create_writer(
            validator=validator, compress=compress)

    def close(self):
        self._fp.close()

    def write(self, url, inner_header=None, http_header=None, data=None, flush=False):
        self._record_writer.write(
            self._fp, url,
            inner_header=inner_header,
            http_header=http_header,
            data=data)

        if flush:
            self._fp.flush()


write = create_writer().write



import sys
from spage_reader import SpageReader, read
from spage_writer import SpageWriter, write


def open_file(name, mode, **kwargs):
    def not_support(name, **kwargs):
        raise ValueError, "mode must be 'r' or 'w'"
    r = {'w': SpageWriter, 'r': SpageReader}.get(mode, not_support)
    return r(name, **kwargs)


__all__ = ['__version__', 'version_info', 'open_file']

import pkgutil
__version__ = pkgutil.get_data(__package__, 'VERSION').decode('ascii').strip()
version_info = tuple(int(v) if v.isdigit() else v
                     for v in __version__.split('.'))
del pkgutil



from collections import OrderedDict
from datetime import datetime

RECORD_TYPES = {"flat", "deleted", "compressed"}


INNER_HEADER_SCHEMA = {
    "type": "object",
    "properties": OrderedDict([
        ("Version", {
            "type": "string",
            "default": "1.2",
        }),
        ("Type", {  # autofill or specify
            "type": "string",
            "enum": RECORD_TYPES,
        }),
        ("Fetch-Time", {  # record store time
            "anyOf": [
                {
                    "type": "datetime",
                },
                {
                    "type": "string",
                    "format": "readable_time"
                },
            ],
            "default": datetime.now,
        }),
        ("Original-Size", {  # data(html) size, autofill
            "type": "number",
        }),
        ("Store-Size", {  # store size, autofill
            "type": "number",
        }),
        ("batchID", {  # batch identity
            "type": "string",
            "minLength": 3,
            "default": '__CHANGE_ME__',
        }),
        ("attach", {
            "type": "string",
        }),
        ("IP-Address", {  # remote host ip
            "type": "string",
            "format": "ipv4",
            "default": "0.0.0.0",
        }),
        ("Spider-Address", {  # spider node identity
            "type": "string",
            "default": "0.0.0.0",
        }),
        ("Digest", {  # can be html md5
            "type": "string",
            "default": "0" * 32,
            "maxLength": 32,
            "minLength": 32,
        }),
        ("User-Agent", {
            "type": "string",
        }),
        ("Fetch-IP", {  # generate page machine ip
            "type": "string",
            "format": "ipv4",
            "default": "0.0.0.0",
        }),
        ("Node-Fetch-Time", {  # real fetch time
            "anyOf": [
                {
                    "type": "datetime",
                },
                {
                    "type": "string",
                    "format": "readable_time"
                },
            ]
        }),
        ("Error-Reason", {
            "type": "string",
            "format": "error_reaseon",
        }),
    ]),
}


META_SCHEMA = {
    "type": "object",
    "properties": {
        "url": {
            "type": "string",
            "format": "url",
        },
        "inner_header": INNER_HEADER_SCHEMA,
        "http_header": {
            "type": "object"
        },
        "data": {
            "type": "string",
        },
    },
    "required": ["url", "inner_header"],
}



import zlib

import pytest
from jsonschema import ValidationError

from os_spage import open_file, read, write


def test_write_invalid_data(tmpdir):
    with tmpdir.as_cwd():
        f = open_file('test', 'w')
        with pytest.raises(ValidationError):
            f.write(url='abc', inner_header={'batchID': 'test'})


def check_inner_header(w_inner_header, r_inner_header):
    if not w_inner_header:
        assert r_inner_header['batchID'] == '__CHANGE_ME__'
        return

    for k, v in w_inner_header.iteritems():
        k = str(k)
        if k not in {'Store-Size', 'Original-Size'}:
            assert r_inner_header[k] == str(v)


def check_http_header(w_http_header, r_http_header):
    if w_http_header is None:
        assert r_http_header == {}
        return
    for k, v in w_http_header.iteritems():
        assert r_http_header[str(k)] == str(v)


def check_data(w_data, r_data):
    if w_data is None:
        assert r_data is None
        return
    assert zlib.decompress(r_data) == w_data


RECORDS = [
    # inner_header, http_header, data
    (None, None, None),
    ({'batchID': 'test'}, {'k1': 'v1'}, 'hello'),
    ({'batchID': 'test'}, {}, 'hello'),
    ({'batchID': 'test'}, {'k1': 'v1'}, None),
    ({}, {'k1': 'v1'}, 'hello'),
    ({}, {'k1': 'v1'}, None),
]


def test_spage_reader_and_writer(tmpdir):
    base_url = "http://www.test.com/"
    filename_prefix = 'test_file_'
    with tmpdir.as_cwd():
        f = open_file(filename_prefix, 'w', roll_size=100)
        idx = 0
        for inner_header, http_header, data in RECORDS:
            url = base_url + str(idx)
            idx += 1
            f.write(url, inner_header=inner_header,
                    http_header=http_header, data=data, flush=True)
        f.close()
        f = open_file(filename_prefix, 'r')
        idx = 0
        for record in f.read():
            inner_header, http_header, data = RECORDS[idx]
            url = base_url + str(idx)
            idx += 1
            assert url == record['url']
            check_inner_header(inner_header, record['inner_header'])
            check_http_header(http_header, record['http_header'])
            check_data(data, record['data'])


def test_genergal_read_and_write(tmpdir):
    base_url = "http://www.test.com/"
    with tmpdir.as_cwd():
        filename = 'test_file'
        f = open(filename, 'w')
        idx = 0
        for inner_header, http_header, data in RECORDS:
            url = base_url + str(idx)
            idx += 1
            write(f, url, inner_header=inner_header,
                  http_header=http_header, data=data)
        f.close()
        f = open(filename, 'r')
        idx = 0
        for record in read(f):
            inner_header, http_header, data = RECORDS[idx]
            url = base_url + str(idx)
            idx += 1
            assert url == record['url']
            check_inner_header(inner_header, record['inner_header'])
            check_http_header(http_header, record['http_header'])
            check_data(data, record['data'])



from setuptools import setup, find_packages


def read(*filenames, **kwargs):
    import io
    from os.path import join, dirname
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(join(dirname(__file__), filename), encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


setup(
    name='os-spage',
    version=read('src/os_spage/VERSION'),
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    license='MIT License',
    description='Read and write spage.',
    long_description=open('README.md').read(),
    author='Ozzy',
    author_email='cfhamlet@gmail.com',
    url='https://github.com/cfhamlet/os-spage',
    zip_safe=False,
    install_requires=['os-rotatefile', 'jsonschema'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
    ])
